# utilities to perform datetime calculations
from datetime import datetime
from pytz import timezone
from datetime import timedelta

def get_clip_name_from_unix_time(source_guid, current_clip_start_time):
    """

    """

    # convert unix time to 
    readable_datetime = datetime.fromtimestamp(int(current_clip_start_time)).strftime('%Y_%m_%d_%H_%M_%S')
    clipname = source_guid + "_" + readable_datetime
    return clipname, readable_datetime

def get_difference_between_times_in_seconds(unix_time1, unix_time2):
    dt1 = datetime.fromtimestamp(int(unix_time1))
    dt2 = datetime.fromtimestamp(int(unix_time2))

    return (dt1-dt2).total_seconds()

def add_interval_to_unix_time(unix_time, interval_in_seconds):
    dt1 = datetime.fromtimestamp(int(unix_time)) + timedelta(0, interval_in_seconds)
    dt1_aware = timezone('US/Pacific').localize(dt1)
    end_time_unix = int(dt1_aware.timestamp())

    return end_time_unix

def get_unix_time_from_datetime_utc(dt_utc):

    dt_aware = timezone('UTC').localize(dt_utc)
    dt_pst = dt_aware.astimezone(timezone('US/Pacific'))

    # convert to PST
    unix_time = int(dt_pst.timestamp())

    return unix_time



#! /usr/bin/env python3
import boto3
from botocore import UNSIGNED
from botocore.config import Config

# Borrowed pagination code from https://alexwlchan.net/2019/07/listing-s3-keys/
def get_all_folders(bucket: str, prefix: str) -> [str]:
    """
    Generate objects in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch objects whose key starts with
        this prefix (optional).
    :param suffix: Only fetch objects whose keys end with
        this suffix (optional).
    """
    s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    paginator = s3.get_paginator("list_objects_v2")

    kwargs = {'Bucket': bucket, 'Prefix': prefix, 'Delimiter': "/"}

    all_keys = []
    # Orcasound buckets are not predictably spaced throughout the time that they've been up some are 2 hours, some hold 24 hours
    # So not making any assumptions

    for page in paginator.paginate(**kwargs):
        try:
            common_prefixes = page["CommonPrefixes"]
            prefixes = [prefix["Prefix"].split("/")[-2] for prefix in common_prefixes]
            all_keys.extend(prefixes)

        except KeyError:
            print("No content returned")
            break

    return all_keys

def get_folders_between_timestamp(bucket_list: [str], start_time: str, end_time: str) -> [int]:
    bucket_list = [int(bucket) for bucket in bucket_list]
    start_index = 0
    end_index = len(bucket_list) - 1
    while int(bucket_list[start_index]) < int(start_time):
        start_index += 1
    while int(bucket_list[end_index]) > int(end_time):
        end_index -= 1
    return bucket_list[start_index-1:end_index + 1]



from . import s3_utils
import boto3
from botocore import UNSIGNED
from botocore.config import Config

import m3u8
import math
from . import scraper
import ffmpeg
import os
from datetime import datetime
from datetime import timedelta
from pytz import timezone
import time
import sys
from pathlib import Path
from . import datetime_utils

def get_readable_clipname(hydrophone_id, cliptime_utc):
    # cliptime is of the form 2020-09-27T00/16/55.677242Z
    cliptime_utc = timezone('UTC').localize(cliptime_utc)
    date = cliptime_utc.astimezone(timezone('US/Pacific'))
    date_format='%Y_%m_%d_%H_%M_%S_%Z'
    clipname = date.strftime(date_format)
    return hydrophone_id + "_" + clipname, date

#TODO: Handle date ranges that don't exist
class DateRangeHLSStream():
    """
    stream_base = 'https://s3-us-west-2.amazonaws.com/streaming-orcasound-net/rpi_orcasound_lab'
    polling_interval = 60 sec
    start_unix_time
    end_unix_time
    wav_dir
    real_time = if False, get data as soon as possible, if true wait for polling interval before pulling
    """

    def __init__(self, stream_base, polling_interval, start_unix_time, end_unix_time, wav_dir, real_time=False):
        """

        """

        # Get all necessary data and create index
        self.stream_base = stream_base
        self.polling_interval_in_seconds = polling_interval
        self.start_unix_time = start_unix_time
        self.end_unix_time = end_unix_time
        self.wav_dir = wav_dir
        self.real_time = real_time
        self.is_end_of_stream = False

        # query the stream base for all m3u8 files between the timestamps

        # split the stream base into bucket and folder
        # eg. 'https://s3-us-west-2.amazonaws.com/streaming-orcasound-net/rpi_orcasound_lab'
        # would be split into s3_bucket = 'streaming-orcasound-net' and folder_name = 'rpi_orcasound_lab'

        bucket_folder = self.stream_base.split("https://s3-us-west-2.amazonaws.com/")[1]
        tokens = bucket_folder.split("/")
        self.s3_bucket = tokens[0]
        self.folder_name = tokens[1]
        prefix = self.folder_name + "/hls/"

        # returns folder names corresponding to epochs, this grows as more data is added, we should probably maintain a list of 
        # hydrophone folders that exist
        all_hydrophone_folders = s3_utils.get_all_folders(self.s3_bucket, prefix=prefix)
        print("Found {} folders in all for hydrophone".format(len(all_hydrophone_folders)))

        self.valid_folders = s3_utils.get_folders_between_timestamp(all_hydrophone_folders, self.start_unix_time, self.end_unix_time)
        print("Found {} folders in date range".format(len(self.valid_folders)))

        self.current_folder_index = 0
        self.current_clip_start_time = self.start_unix_time

    def get_next_clip(self, current_clip_name = None):
        """

        """

        # Get current folder
        current_folder = int(self.valid_folders[self.current_folder_index])
        clipname, clip_start_time = datetime_utils.get_clip_name_from_unix_time(self.folder_name.replace("_", "-"), self.current_clip_start_time)

        # if real_time execution mode is specified
        if self.real_time:
            # sleep till enough time has elapsed

            now = datetime.utcnow()
            time_to_sleep = (current_clip_name-now).total_seconds()

            if time_to_sleep < 0:
                print("Issue with timing")

            if time_to_sleep > 0:
                time.sleep(time_to_sleep)
        
        # read in current m3u8 file
        # stream_url for the current AWS folder
        stream_url = "{}/hls/{}/live.m3u8".format(
            (self.stream_base), (current_folder))
        stream_obj = m3u8.load(stream_url)
        num_total_segments = len(stream_obj.segments)
        num_segments_in_wav_duration = math.ceil(self.polling_interval_in_seconds/stream_obj.target_duration)

        # calculate the start index by computing the current time - start of current folder
        segment_start_index = math.ceil(datetime_utils.get_difference_between_times_in_seconds(self.current_clip_start_time, current_folder)/stream_obj.target_duration)
        segment_end_index = segment_start_index + num_segments_in_wav_duration

        if segment_end_index > num_total_segments:
            # move to the next folder and increment the current_clip_start_time to the new
            self.current_folder_index += 1
            self.current_clip_start_time = self.valid_folders[self.current_folder_index]
            return None, None, None

        # Can get the whole segment so update the clip_start_time for the next clip
        # We do this before we actually do the pulling in case there is a problem with this clip
        self.current_clip_start_time = datetime_utils.add_interval_to_unix_time(self.current_clip_start_time, self.polling_interval_in_seconds)

        # Create tmp path to hold .ts segments
        tmp_path = "tmp_path"
        os.makedirs(tmp_path,exist_ok=True)

        file_names = []
        for i in range(segment_start_index, segment_end_index):
            audio_segment = stream_obj.segments[i]
            base_path = audio_segment.base_uri
            file_name = audio_segment.uri
            audio_url = base_path + file_name
            try:
                scraper.download_from_url(audio_url,tmp_path)
                file_names.append(file_name)
            except Exception:
                print("Skipping",audio_url,": error.")

        # concatentate all .ts files with ffmpeg
        hls_file = (clipname+".ts")
        audio_file = (clipname+".wav")
        wav_file_path = os.path.join(self.wav_dir, audio_file)
        filenames_str = " ".join(file_names)
        concat_ts_cmd = "cd {tp} && cat {fstr} > {hls_file}".format(tp=tmp_path, fstr=filenames_str, hls_file=hls_file)
        os.system(concat_ts_cmd)
        
        # read the concatenated .ts and write to wav
        stream = ffmpeg.input(os.path.join(tmp_path, Path(hls_file)))
        stream = ffmpeg.output(stream, wav_file_path)
        ffmpeg.run(stream, quiet=False)

        # clear the tmp_path
        os.system(f'rm -rf {tmp_path}')

        # If we're in demo mode, we need to fake timestamps to make it seem like the date range is real-time
        if current_clip_name:
            clipname, _ = get_readable_clipname(self.folder_name.replace("_", "-"), current_clip_name)

            # rename wav file
            full_new_clip_path = os.path.join(self.wav_dir, clipname+".wav")
            os.rename(wav_file_path, full_new_clip_path)
            wav_file_path = full_new_clip_path

            # change clip_start_time - this has to be in UTC so that the email can be in PDT
            clip_start_time = current_clip_name.isoformat() + "Z"

        # Get new index
        return wav_file_path, clip_start_time, current_clip_name

    def is_stream_over(self):
        # returns true or false based on whether the stream is over
        return int(self.current_clip_start_time) >= int(self.end_unix_time)


import os, sys, json, glob
import requests, bs4, urllib
import tqdm
import pandas as pd
from os.path import dirname
from urllib.parse import urljoin

class TqdmUpTo(tqdm.tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""
    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def _geturlsoup(url):
    response = requests.get(url)
    return bs4.BeautifulSoup(response.text,"html.parser")

def _get_table_rows(url,occurence=0):
    response = requests.get(url)
    soup = bs4.BeautifulSoup(response.text,"html.parser")
    return soup.findAll('table')[occurence].findAll('tr')

def _get_urls_from_select_button(soup,baseurl):
    url_dict = {}
    options = soup.findAll('option')
    for option in options:
        key = option.string.strip().lower()
        value = '/'.join([baseurl,option['value']])
        url_dict[key]=value
    if "select" in url_dict: del url_dict["select"]
    return url_dict

def parse_database_page_to_tsv(url,tsvfile):
    baseurl = '/'.join(url.split('/',3)[:-1])
    rows = _get_table_rows(url)
    getstr = lambda x: x.string.strip() if x.string is not None else ""

    with open(tsvfile,'w') as f:
        f.write("name\tlocation\tdate\taudio\tmetadata\n")
        for row in rows[1:]: # skip title row
            contents = row.findAll('td')
            name, loc, date, audio, metadata = getstr(contents[1]), \
                getstr(contents[2]),\
                getstr(contents[3]),\
                urljoin(baseurl,contents[4].a['href']),\
                '/'.join([
                    url.rsplit('/',1)[0],
                    contents[5].a['href'].split('(')[1].split(')')[0].replace('\'','')
                ])
            f.write('\t'.join([name,loc,date,audio,metadata])+'\n')
    print("Parsed links for",len(rows),"rows to file:",tsvfile)

def select_and_get_urls(main_url,common_name): # for fun could be done recursively
    main_soup = _geturlsoup(main_url)
    select_buttons = main_soup.findAll('select')
    # select name get initial page url
    url_dict = _get_urls_from_select_button(select_buttons[0],dirname(main_url))
    select_soup = _geturlsoup(url_dict[common_name])
    select_buttons = select_soup.findAll('select')
    return _get_urls_from_select_button(select_buttons[2],dirname(main_url))

def get_metadata(mtd_url):
    metadata = {}
    rows = _get_table_rows(mtd_url,1)
    for row in rows[1:]:
        contents = row.findAll('td')
        metadata[contents[0].string.replace(':','')] = contents[1].string
    return metadata

def where_are_the_whales(main_url,save_dir):
    soup = _geturlsoup(main_url).findAll('select')[0] # main select drop down
    ud = _get_urls_from_select_button(soup,dirname(main_url))
    common_names = [ k for k in ud.keys() if 'whale' in k ] # just get all whale species

    for common_name in common_names:
        url_years = select_and_get_urls(main_url,common_name)
        filedir = os.path.join(save_dir,common_name)
        os.makedirs(filedir,exist_ok=True)
        for year, url in url_years.items():
            parse_database_page_to_tsv(
                url,
                "{}_{}.tsv".format(os.path.join(filedir,common_name),str(year))
                )

num_lines = lambda x: sum(1 for line in open(x))
def fetch_all_metadata(save_dir,whale):
    whales = os.listdir(save_dir)

    metadata_json = {}
    for tsv_file in glob.glob(os.path.join(save_dir,whale,"*.tsv")):
        with open(tsv_file,'r') as f:
            next(f) # skip header row
            for line in tqdm.tqdm(f,total=num_lines(tsv_file)):
                mtd_url = line.split('\t')[-1].strip()
                metadata_json[mtd_url] = get_metadata(mtd_url)

    with open(os.path.join(save_dir,whale,"metadata.json"),'w') as f:
        json.dump(metadata_json,f)

def download_from_url(dl_url,dl_dir):
    # download only if not already exists
    file_name = os.path.basename(dl_url)
    dl_path = os.path.join(dl_dir,file_name)
    if os.path.isfile(dl_path):
        print("Skipping",file_name,"as it already exists.")
    else:
        print("Downloading",file_name)
        with TqdmUpTo(unit='B', unit_scale=True, miniters=1,
                    desc=dl_url.split('/')[-1]) as t:  # all optional kwargs
            urllib.request.urlretrieve(dl_url, filename=dl_path,
                            reporthook=t.update_to)

def download_all_cuts(save_dir,whale,wav_dir):
    cuts_tsv = os.path.join(save_dir,whale,"allcuts.tsv")
    wav_dir = os.path.join("./data/wavcut",whale)

    os.makedirs(wav_dir,exist_ok=True)
    df = pd.read_csv(cuts_tsv,sep='\t')

    # get file
    for dl_url in tqdm.tqdm(df.audio):
        download_from_url(dl_url,wav_dir)


if __name__ == "__main__":
    main_url = "https://cis.whoi.edu/science/B/whalesounds/fullCuts.cfm"
    save_dir = "./data/tsv"
    whale = "killer whale"
    cuts_tsv = os.path.join(save_dir,whale,"allcuts.tsv")
    wav_dir = os.path.join("./data/wavcut",whale)



# HLSStream class 
import urllib.request
import os
import m3u8
import math
import ffmpeg #ffmpeg-python
from . import scraper
from datetime import datetime
from datetime import timedelta

import time
from pytz import timezone
from . import datetime_utils
from pathlib import Path

import boto3
from botocore import UNSIGNED
from botocore.config import Config

def get_readable_clipname(hydrophone_id, cliptime_utc):
    # cliptime is of the form 2020-09-27T00/16/55.677242Z
    cliptime_utc = timezone('UTC').localize(cliptime_utc)
    date = cliptime_utc.astimezone(timezone('US/Pacific'))
    date_format='%Y_%m_%d_%H_%M_%S_%Z'
    clipname = date.strftime(date_format)
    return hydrophone_id + "_" + clipname, date

#TODO (@prgogia) Handle errors due to rebooting of hydrophone
class HLSStream():
    """
    stream_base = 'https://s3-us-west-2.amazonaws.com/streaming-orcasound-net/rpi_orcasound_lab'
    polling_interval = 60 sec
    """

    def __init__(self, stream_base, polling_interval, wav_dir):
        self.stream_base = stream_base
        self.polling_interval = polling_interval
        self.wav_dir = wav_dir
        bucket_folder = self.stream_base.split("https://s3-us-west-2.amazonaws.com/")[1]
        tokens = bucket_folder.split("/")
        self.s3_bucket = tokens[0]
        self.hydrophone_id = tokens[1]

    # this function grabs audio from last_end_time to 
    def get_next_clip(self, current_clip_end_time):

        # if current time < current_clip_end_time, sleep for the difference
        now = datetime.utcnow()

        # the extra 10 seconds to sleep is to download the last .ts segment properly
        time_to_sleep = (current_clip_end_time-now).total_seconds() + 10

        if time_to_sleep < 10:
            # This implies that we took more than polling interval seconds to do our processing
            # Ideally, we should set current_clip_end_time = now at this point
            current_clip_end_time = now
            time.sleep(10)

            print("Fell behind now so fast forwarding current_clip_end_time")
        else:
            # In steady state, time_to_sleep is ~41 seconds
            # the very first time, it's ~20 seconds
            time.sleep(time_to_sleep)

        current_clip_start_time = current_clip_end_time - timedelta(0,60)
        clip_start_time = current_clip_start_time.isoformat() + "Z"
        clipname, _ = get_readable_clipname(self.hydrophone_id, current_clip_start_time)

        # get latest AWS bucket
        print("Listening to location {loc}".format(loc=self.stream_base))
        latest = f"{self.stream_base}/latest.txt"
        stream_id = urllib.request.urlopen(
            latest).read().decode("utf-8").replace("\n", "")

        # stream_url for the current AWS bucket
        stream_url = "{}/hls/{}/live.m3u8".format(
            (self.stream_base), (stream_id))
        
        # check if m3u8 file exists
        s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        prefix = "{}/hls/{}/live.m3u8".format(self.hydrophone_id, stream_id)
        objs = s3.list_objects_v2(Bucket=self.s3_bucket, Prefix=prefix, MaxKeys=1, Delimiter="/")
        
        # if it does not, exit
        if objs["KeyCount"] == 0:
            print(".m3u8 file does not exist, will retry after some time")
            return None, None, current_clip_end_time
        
        assert objs["KeyCount"] == 1

        # .m3u8 file exists so load it
        stream_obj = m3u8.load(stream_url)
        num_total_segments = len(stream_obj.segments)
        num_segments_in_wav_duration = math.ceil(self.polling_interval/stream_obj.target_duration)

        # calculate the start index by computing the current time - start of current folder
        current_clip_end_time_unix_pst = datetime_utils.get_unix_time_from_datetime_utc(current_clip_end_time)
        time_since_folder_start = datetime_utils.get_difference_between_times_in_seconds(current_clip_end_time_unix_pst, stream_id)
        if time_since_folder_start < self.polling_interval + 20:
            # This implies that possibly a new folder was created 
            # and we do not have enough data for a 1 minute clip + 20 second buffer 
            # we exit and try again after hls polling interval
            None, None, current_clip_end_time
        
        min_num_total_segments_required = math.ceil(time_since_folder_start/stream_obj.target_duration)
        segment_start_index = min_num_total_segments_required - num_segments_in_wav_duration + 1
        segment_end_index = segment_start_index + num_segments_in_wav_duration

        if segment_end_index > num_total_segments:
            return None, None, current_clip_end_time

        # Create tmp path to hold .ts segments
        tmp_path = "tmp_path"
        os.makedirs(tmp_path,exist_ok=True)

        file_names = []
        for i in range(segment_start_index, segment_end_index):
            audio_segment = stream_obj.segments[i]
            base_path = audio_segment.base_uri
            file_name = audio_segment.uri
            audio_url = base_path + file_name
            try:
                scraper.download_from_url(audio_url,tmp_path)
                file_names.append(file_name)
            except Exception:
                print("Skipping",audio_url,": error.")

        # concatentate all .ts files with ffmpeg
        hls_file = (clipname+".ts")
        audio_file = (clipname+".wav")
        wav_file_path = os.path.join(self.wav_dir, audio_file)
        filenames_str = " ".join(file_names)
        concat_ts_cmd = "cd {tp} && cat {fstr} > {hls_file}".format(tp=tmp_path, fstr=filenames_str, hls_file=hls_file)
        os.system(concat_ts_cmd)
        
        # read the concatenated .ts and write to wav
        stream = ffmpeg.input(os.path.join(tmp_path, Path(hls_file)))
        stream = ffmpeg.output(stream, wav_file_path)
        ffmpeg.run(stream, quiet=True)

        # clear the tmp_path
        os.system(f'rm -rf {tmp_path}')

        return wav_file_path, clip_start_time, current_clip_end_time

    def is_stream_over(self):
        # returns true or false based on whether the stream is over
        # for a live stream, the stream is never over
        return False
